package cmd

import (
	"bytes"
	"context"
	"errors"
	"fmt"
	"os"
	"syscall"

	"github.com/jmoiron/sqlx"
	"golang.org/x/crypto/bcrypt"
	"golang.org/x/term"
	"orus.io/orus-io/go-orusapi/database"

	"orus.io/orus-io/rednerd/models"
	"orus.io/orus-io/rednerd/utils"
)

// ErrPasswordMismatch is returned if the password aren't equal.
var ErrPasswordMismatch = errors.New("password mismatch")

// AdminPasswordCmd is the 'admin-user' command.
type AdminPasswordCmd struct {
	options *Options `no-flag:"t"`
}

// Execute reset the admin user password.
func (cmd *AdminPasswordCmd) Execute(args []string) error {
	ctx, cancel := utils.CtxWithSignal(context.Background(), os.Interrupt, syscall.SIGQUIT, syscall.SIGTERM)
	defer cancel()

	if len(args) > 1 {
		return fmt.Errorf("invalid number of args. Expects 1 or 0, got %d", len(args))
	}

	db, err := sqlx.Open("postgres", cmd.options.Dsn)
	if err != nil {
		return err
	}

	defer func() {
		if err := db.Close(); err != nil {
			log.Err(err).Msg("could not close database connection properly")
		}
	}()

	if err := db.Ping(); err != nil {
		return err
	}

	tx, err := database.Begin(ctx, db)
	if err != nil {
		return err
	}

	defer tx.RollbackIfOpened(log)

	var password []byte

	if len(args) == 0 {
		fmt.Print("Type the new admin password: ")

		p, err := term.ReadPassword(int(os.Stdin.Fd()))
		if err != nil {
			return err
		}

		fmt.Println()

		fmt.Print("Please confirm: ")

		passwordConfirm, err := term.ReadPassword(int(os.Stdin.Fd()))
		if err != nil {
			return err
		}

		fmt.Println()

		if !bytes.Equal(p, passwordConfirm) {
			return ErrPasswordMismatch
		}

		password = p
	} else if len(args) == 1 {
		password = []byte(args[0])
	}

	if err := SetAdminPassword(ctx, tx.Tx, password); err != nil {
		return err
	}

	return tx.Commit()
}

// SetAdminPassword resets the admin account and its password.
func SetAdminPassword(ctx context.Context, tx *sqlx.Tx, password []byte) error {
	// Salt and hash the password using the bcrypt algorithm
	// The second argument is the cost of hashing, which we arbitrarily set as 8
	// (this value can be more or less, depending on the computing power you wish to utilize)
	hashedPassword, err := bcrypt.GenerateFromPassword(password, 10)
	if err != nil {
		return err
	}

	var (
		account = models.AccountDB{
			Name:   "admin",
			IsUser: true,
		}
		user = models.UserDB{
			Userinfo: models.Userinfo{
				Username: "admin",
				IsAdmin:  true,
			},
			Password: string(hashedPassword),
		}
	)

	h := database.NewSQLHelper(ctx, tx, log)

	if err := h.Upsert(account); err != nil {
		return err
	}

	if err := h.Upsert(user); err != nil {
		return err
	}

	return nil
}
